/* JSGalBuilder: (c) 2009 by Andreas S. Krebs
 * 
 * This file is part of JSGalBuilder.
 *
 * JSGalBuilder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JSGalBuilder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JSGalBuilder.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sberk.jsgal.datamodel;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.xml.parsers.ParserConfigurationException;

import net.sberk.jsgal.datamodel.dist.JSGalDist;
import net.sberk.util.MiscUtil;
import net.sberk.util.XMLUtil;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

@SuppressWarnings("nls")
public class XHTMLPage
{
	private final static String		PUBLICID	="-//W3C//DTD XHTML 1.0 Strict//EN";
	private final static String		SYSTEMID	="http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd";
	 
	private final static String		SCRIPT		="script";
	private final static String		TITLE		="title";
	private final static String		LINK		="link";
	private final static String		ID			="id";
	private final static String		GALLERIES	="galleries";
	private final static String		HREF		="href";
	private final static String		META		="meta";
	private final static String		NAME		="name";
	private final static String		VERSION		="version";
	private final static String		CONTENT		="content";

	private final static String		VERSION_PREFIX	="JSGal v";
	
	private File				xhtmlFile;
	private String				title;
	private File				galsFile;
	private GalleryCollection	galCollection;
	private String				version;
	private float				versionNumber;
	
	public XHTMLPage(File xhtmlFile, String title, File galsFile, String version)
	{
		this(xhtmlFile, title, galsFile, version, new GalleryCollection());
	}
	
	public XHTMLPage(File xhtmlFile, String title, File galsFile, String version, GalleryCollection gc)
	{
		this.xhtmlFile=xhtmlFile;
		this.title=title;
		this.galsFile=galsFile;
		this.version=version;
		this.galCollection=gc;
		
		updateVersionInfo();
	}
	
	private void updateVersionInfo()
	{
		float	vn=-1;
		
		if(version==null)
		{
			version="UNKNOWN";
			vn=0;
		}
		else if(!version.startsWith(VERSION_PREFIX))
			version="INVALID ("+version+")";
		else
		{
			try
			{
				vn=Float.parseFloat(version.substring(VERSION_PREFIX.length()));
			}
			catch(Exception e)
			{
				version="INVALID ("+version+")";
			}
		}

		versionNumber=vn;
	}

	public void create() throws Exception
	{
		if(xhtmlFile.exists())
			throw new IOException("XHTML page alread exists: "+xhtmlFile);

		if(galsFile.exists())
			throw new IOException("Gallieres XML file alread exists: "+galsFile);
		
		save(JSGalDist.getNewIndexDocument(), false);
	}
	
	public void save() throws Exception
	{
		save(XMLUtil.parseXML(xhtmlFile, false), false);
	}
	
	public File upgrade() throws Exception
	{
		// backup the current XHTML page
		File	bakFile=MiscUtil.backupFile(xhtmlFile);
		
		// now save new page using the default document from the dist
		save(JSGalDist.getNewIndexDocument(), true);
		
		return bakFile;
	}
	
	private void save(Document doc, boolean upgrade) throws Exception
	{
		// update the title and galleries link
		updateDocumentTags(doc);
		
		// first save indexFile 
		// Note: in order to prevent problems with IE6, the xml declaration should be ommited
		XMLUtil.writeXML(doc, false, true, PUBLICID, SYSTEMID, new FileOutputStream(xhtmlFile));

		// now (re)create the files from the jsgal distribution if missing or new
		if(upgrade)
			JSGalDist.upgradeJSGAL(this);
		else
			JSGalDist.generateJSGal(this);
		
		// finally save the gallery collection
		galCollection.writeXML(galsFile);
	}
	
	public File getXhtmlFile() 
	{
		return xhtmlFile;
	}

	public String getJSGalVersion() 
	{
		return version;
	}
	
	public float getJSGalVersionNumber() 
	{
		return versionNumber;
	}

	public String getTitle() 
	{
		return title;
	}

	public void setTitle(String title)
	{
		this.title=title;
	}
	
	public File getGalleriesFile() 
	{
		return galsFile;
	}

	public GalleryCollection getGalCollection() 
	{
		return galCollection;
	}
	
	public static XHTMLPage parseXHTML(File file) throws IOException, ParserConfigurationException, SAXException, IllegalArgumentException
	{
		File f=file.getCanonicalFile();	// make sure we have the complete path
		
		Document	doc=XMLUtil.parseXML(f, false);
		
		// we don't really parse the page only get the tags we need
		String		title, gals, version;
		
		title=getTitleFromDoc(doc);
		gals=getGalleriesFromDoc(doc);
		version=getVersionFromDoc(doc);
		
		if(title==null)
			title=f.getName();
		
		if(gals==null || gals.length()<=0)
			throw new IllegalArgumentException("No link tag with id='galleries' found!");
		
		File				gf=new File(f.getParentFile(), gals);
		GalleryCollection	gc=GalleryCollection.parseXML(gf);
		
		return new XHTMLPage(f, title, gf, version, gc);
	}
	
	public static XHTMLPage parseOnlyUsedTags(Document doc)
	{
		// we don't really parse the page only get the tags we need
		String		title, gals, version;
		
		title=getTitleFromDoc(doc);
		gals=getGalleriesFromDoc(doc);
		version=getVersionFromDoc(doc);
		
		File				gf=new File(gals);
		
		return new XHTMLPage(null, title, gf, version);
	}
	
	private static String getTitleFromDoc(Document doc)
	{
		NodeList	nl=doc.getElementsByTagName(TITLE);
		for(int i=0; i<nl.getLength(); i++)
		{
			Node	nd=nl.item(i);
			
			if(nd instanceof Element)
			{
				Element	e=(Element)nd;
				
				return XMLUtil.getElementText(e);
			}
		}
		
		return null;
	}
	
	private static String getGalleriesFromDoc(Document doc)
	{
		NodeList	nl=doc.getElementsByTagName(LINK);
		for(int i=0; i<nl.getLength(); i++)
		{
			Node	nd=nl.item(i);
			
			if(nd instanceof Element)
			{
				Element	e=(Element)nd;
				
				if(GALLERIES.equals(e.getAttribute(ID)))
				{
					return e.getAttribute(HREF);
				}
			}
		}
		
		return null;
	}
	
	private static String getVersionFromDoc(Document doc)
	{
		NodeList	nl=doc.getElementsByTagName(META);
		for(int i=0; i<nl.getLength(); i++)
		{
			Node	nd=nl.item(i);
			
			if(nd instanceof Element)
			{
				Element	e=(Element)nd;
				
				if(VERSION.equals(e.getAttribute(NAME)))
				{
					return e.getAttribute(CONTENT);
				}
			}
		}
		
		return null;
	}
	
	private void updateDocumentTags(Document doc) throws IOException
	{
		// now update the tags we need
		NodeList	nl=doc.getElementsByTagName(TITLE);
		for(int i=0; i<nl.getLength(); i++)
		{
			Node	nd=nl.item(i);
			
			if(nd instanceof Element)
			{
				Element	e=(Element)nd;
				
				// remove old title if present
				while(e.getFirstChild()!=null)
				{
					e.removeChild(e.getFirstChild());
				}
				
				// now create new text node
				e.appendChild(doc.createTextNode(title));
			}
		}

		nl=doc.getElementsByTagName(LINK);
		for(int i=0; i<nl.getLength(); i++)
		{
			Node	nd=nl.item(i);
			
			if(nd instanceof Element)
			{
				Element	e=(Element)nd;
				
				if(GALLERIES.equals(e.getAttribute(ID)))
				{
					e.setAttribute(HREF, MiscUtil.getRelativePath(xhtmlFile, galsFile));
				}
			}
		}	
		
		// to prevent problems with multiple browser the script tag has to contain some dummy strings
		nl=doc.getElementsByTagName(SCRIPT);
		for(int i=0; i<nl.getLength(); i++)
		{
			Node	nd=nl.item(i);
			
			if(nd instanceof Element)
			{
				Element	e=(Element)nd;
				
				if(e.getChildNodes().getLength()<=0)
				{
					// no child nodes so we add a dummy comment
					e.appendChild(doc.createComment("NO SELF CLOSE"));
				}
			}
		}
	}
}