/* JSGalBuilder: (c) 2009 by Andreas S. Krebs
 * 
 * This file is part of JSGalBuilder.
 *
 * JSGalBuilder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JSGalBuilder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JSGalBuilder.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sberk.jsgal.gui;

import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Transparency;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.io.File;
import java.io.IOException;
import java.text.MessageFormat;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.stream.ImageOutputStream;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileNameExtensionFilter;

import net.sberk.jsgal.datamodel.Gallery;
import net.sberk.jsgal.gui.ImageHelper.GenerationOptions.TimeStampUsage;
import net.sberk.util.ExifExtractor;
import net.sberk.util.MiscUtil;
import net.sberk.util.ExifExtractor.ExifMetaData;

class ImageHelper
{
	private static JFileChooser	fileChooser;
	protected static File[] selectImages(Component parent, File bf, String title, String okText[])
	{
		if(fileChooser==null)
		{
			fileChooser=new JFileChooser();

			fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
			fileChooser.setMultiSelectionEnabled(true);
			
			// create FileFiler
			String	exts[]=ImageIO.getReaderFileSuffixes();		// this returns only the main types
			StringBuilder	sb=new StringBuilder();
			for(int i=0; i<exts.length; i++)
			{
				if(i>0)
					sb.append(',');
				
				sb.append("*.").append(exts[i]); //$NON-NLS-1$
			}
			String	name=MessageFormat.format(I18N.IMAGE_FILEFILTERNAME, sb.toString());
			// getReaderFormatNames returns all upper/lower casing versions of the exts
			FileNameExtensionFilter	filt=new FileNameExtensionFilter(name, ImageIO.getReaderFormatNames());
			
			fileChooser.addChoosableFileFilter(filt);
			fileChooser.setFileFilter(filt);
						
			if(bf!=null)
				fileChooser.setCurrentDirectory(bf.getParentFile());
		}

		fileChooser.setDialogTitle(title);
		fileChooser.setApproveButtonText(okText[0]);
		fileChooser.setApproveButtonMnemonic(okText[1].charAt(0));
		fileChooser.setApproveButtonToolTipText(okText[2]);
		
		fileChooser.setSelectedFile(null);
		
		int rv=fileChooser.showDialog(parent, null);

		return (rv==JFileChooser.APPROVE_OPTION) ? fileChooser.getSelectedFiles() : null;
	}
	
	protected static void addImageToGallery(File f, int rot, GalleryPanel gp, GenerationOptions opts) throws Exception
	{
		String			ext=opts.format;
		File			galFile=gp.getGalleryFile();
		Gallery			g=gp.getGallery();
		File			thumbDir=null, imgDir=MiscUtil.getRelativeFile(galFile, g.getImageDir());
		ExifMetaData	md=null;
		int				rotate=rot;
		
		// we use format as ext except for jpeg
		if(ext.equals("jpeg")) //$NON-NLS-1$
			ext="jpg"; //$NON-NLS-1$
		
		// only create thumbs if we have a seperate dir
		if(!g.getThumbsDir().equals(g.getImageDir()))
			thumbDir=MiscUtil.getRelativeFile(galFile, g.getThumbsDir());
		
		// first read the image
		BufferedImage	bi=ImageIO.read(f);

		String	fn=f.getName();
		if(!fn.endsWith('.'+ext))	// different format
			fn=MiscUtil.changeFileNameExt(fn, ext);
		
		// get rotation from exif if requested
		if(rotate==ROTATE_AUTO)
		{
			md=ExifExtractor.getMetaData(f);
			
			rotate=getRotationFromExif(md.Orientation);
		}
		
		// now create the thumb and the original
		if(thumbDir!=null)
		{
			File	dest=MiscUtil.getRelativeFile(thumbDir, fn);
			
			createScaledImage(dest, bi, g.getThumbWidth(), g.getThumbHeight(), rotate, g.getDefScale(), opts);
		}
		
		// now create the image
		File	dest=MiscUtil.getRelativeFile(imgDir, fn);
		String	scale=createScaledImage(dest, bi, g.getImageWidth(), g.getImageHeight(), rotate, g.getDefScale(), opts);
		
		String	name=fn;
		String	descr=null;

		if(opts.useTimeStampAs!=TimeStampUsage.NONE)
		{
			// now get the meta data if not yet extracted
			if(md==null)
				md=ExifExtractor.getMetaData(f);

			if(opts.useTimeStampAs!=TimeStampUsage.NAME)
				name=md.DateTime;
			else if(opts.useTimeStampAs!=TimeStampUsage.DESCRIPTION)
				descr=md.DateTime;
		}
		
		// finally create the Image and add it to the gallery
		Gallery.Image	gi=new Gallery.Image(fn, name, fn, scale, descr);
		g.addImage(gi);
		gp.new ImageUserData(gi);
	}
	
	private static String createScaledImage(File dest, BufferedImage bi, int targetWidth, int targetHeight, int rotate, String defScale, GenerationOptions opts) throws IOException
	{	
		if(!opts.overwriteExisting && dest.exists())
			throw new IOException(I18NMsg.MSG_IMAGE_FILE_ALREADY_EXISTS.fmt(dest.getAbsolutePath())); 
		
		BufferedImage	img;
		int				w=bi.getWidth();
		int				h=bi.getHeight();
		String			scale=getScaling(w, h, targetWidth, targetHeight, defScale);
		
		// is scaling needed
		if(opts.scaleIncreaseSize || w>targetWidth || h>targetHeight)
			img=getScaledImage(bi, targetWidth, targetHeight, scale, true, rotate);
		else
			img=bi;
				
		saveImage(dest, img, opts);
		
		return scale;
	}
	
	private static void saveImage(File dest, BufferedImage bi, GenerationOptions opts) throws IOException
	{
		ImageWriter		writer=ImageIO.getImageWritersByFormatName(opts.format).next();
		ImageWriteParam	param=writer.getDefaultWriteParam();
		
		try
		{
			param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
			param.setCompressionQuality(opts.quality);
		}
		catch(Exception e) 
		{
			// simply ignore
		}
		
		try
		{
			param.setProgressiveMode(opts.saveProgressive ? ImageWriteParam.MODE_DEFAULT : ImageWriteParam.MODE_DISABLED);
		}
		catch(Exception e) 
		{
			// simply ignore
		}
		
		if(opts.format.equals("jpeg")) //$NON-NLS-1$
		{
			// set color model to take care of saving PNG source
			param.setSourceBands(new int[] {0,1,2});
			
			ColorModel	cm = new DirectColorModel(24, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x0);
			param.setDestinationType(new ImageTypeSpecifier(cm, cm.createCompatibleSampleModel(1,1)));
		}
		
		ImageOutputStream	os=ImageIO.createImageOutputStream(dest);
		
		writer.setOutput(os);
		writer.write(null, new IIOImage(bi, null, null),param);
		
		writer.dispose();
		os.close();
	}
	
	public static String getScaling(int srcW, int srcH, int destW, int destH, String defScale)
	{
		// check if the defScale works
		if(MiscUtil.equals(Gallery.SCALE_HORIZ, defScale))
		{
			int	sh=(srcH * destW) / srcW;
			
			return (sh<=destH) ? Gallery.SCALE_HORIZ : Gallery.SCALE_VERT; 
		}

		// now check for vertical scaling
		int	sw=(srcW * destH) / srcH;
		return (sw<=destW) ? Gallery.SCALE_VERT : Gallery.SCALE_HORIZ; 
	}
	
	private final static double	DEGREES[] = { -1d, 0d, 90d, 180d, 270d };
	public final static int	ROTATE_AUTO	=0;
	public final static int	ROTATE_0	=1;
	public final static int	ROTATE_90	=2;
	public final static int	ROTATE_180	=3;
	public final static int	ROTATE_270	=4;
	
	public static BufferedImage getScaledImage(BufferedImage bi, int targetW, int targetH, String scale, boolean highQuality, int rotate)
	{
		BufferedImage 	img=bi;
		Object			hint=highQuality ? RenderingHints.VALUE_INTERPOLATION_BICUBIC : RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR;
		int				type=(img.getTransparency()==Transparency.OPAQUE) ? BufferedImage.TYPE_INT_RGB : BufferedImage.TYPE_INT_ARGB;
		boolean			multistep=highQuality;
		int				tw, th;
		
		// if this is the last step and rotating is requeste - do it
		if(rotate>ROTATE_0)
		{
			int	w=img.getWidth(), h=img.getHeight();
			int	iw, ih;
			
			if(rotate==ROTATE_90 || rotate==ROTATE_270)
			{
				iw=h;
				ih=w;
			}
			else
			{
				iw=w;
				ih=h;
			}

			BufferedImage	ti=new BufferedImage(iw, ih, type);
			Graphics2D		g2d=ti.createGraphics();
			
			g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, hint);
			
			g2d.translate(iw/2, ih/2);
			g2d.rotate(Math.toRadians(DEGREES[rotate]));
			g2d.drawImage(img, -w/2, -h/2, w, h, null);
			g2d.dispose();

			img=ti;			
		}
		
		// calculate scaled size
		if(MiscUtil.equals(Gallery.SCALE_BOTH, scale))
		{
			tw=targetW;
			th=targetH;
		}
		else if(MiscUtil.equals(Gallery.SCALE_HORIZ, scale))
		{
			tw=targetW;
			th=(targetW*img.getHeight())/img.getWidth();
		}
		else
		{
			th=targetH;
			tw=(targetH*img.getWidth())/img.getHeight();
		}
			
		// a final validation whether scaling is needed at all
		if(tw==img.getWidth() && th==img.getHeight())
			return img;
		
		// now prepare for scaling
		int 	w=multistep ? img.getWidth() : tw;
		int		h=multistep ? img.getHeight() : th;

		BufferedImage	scImg=img;
		do 
		{
			if(multistep && w>tw)
				w/=2;
			if(w<tw)
				w=tw;

			if(multistep && h>th)
				h/=2;
			if(h<th)
				h=th;
			
			// since the quality of AffineTransform is really bad for scaling down
			// we use instead drawImage with a multistep approach which seems to 
			// create reasonable good scaled down images
			BufferedImage	ti=new BufferedImage(w, h, type);
			Graphics2D		g2d=ti.createGraphics();
			
			g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, hint);
			g2d.drawImage(scImg, 0, 0, w, h, null);
			g2d.dispose();

			scImg=ti;
		}while(w!=tw || h!=th);
		
		return scImg;
	}
	
	protected static class GenerationOptions
	{
		public String			format="jpeg"; //$NON-NLS-1$
		public float			quality=0.7f;
		public boolean			saveProgressive=true;
		public boolean			overwriteExisting=false;
		public boolean			scaleIncreaseSize=false;
		public TimeStampUsage	useTimeStampAs=TimeStampUsage.NAME;
		
		public enum TimeStampUsage { NONE, NAME, DESCRIPTION }
	}
	
	public static int getRotationFromExif(int exifOrient)
	{
		if(exifOrient==6)
			return ROTATE_90;
		else if(exifOrient==3)
			return ROTATE_180;
		else if(exifOrient==8)
			return ROTATE_270;
		
		return ROTATE_0;	// do nothing as default
	}
}