/* JSGalBuilder: (c) 2009 by Andreas S. Krebs
 * 
 * This file is part of JSGalBuilder.
 *
 * JSGalBuilder is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JSGalBuilder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JSGalBuilder.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sberk.jsgal.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.util.Vector;

import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultCellEditor;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.WindowConstants;
import javax.swing.TransferHandler.DropLocation;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import net.sberk.jsgal.gui.I18N.NamedAction;
import net.sberk.jsgal.gui.ImageHelper.GenerationOptions;
import net.sberk.jsgal.gui.ImageHelper.GenerationOptions.TimeStampUsage;
import net.sberk.util.MiscUtil;

@SuppressWarnings({ "serial" })
class ImageImportDialog extends JDialog implements Runnable, ListSelectionListener, TableModelListener
{
	private final static String	IMAGE_OUTPUT_FORMATS[] = { "jpeg", "png", "gif" };  //$NON-NLS-1$//$NON-NLS-2$ //$NON-NLS-3$
	
	private Action	ACTION_FILE_ADD;
	private Action	ACTION_FILE_REMOVE;
	private Action	ACTION_START_CONVERSION;
	private Action	ACTION_STOP_CONVERSION;
	private Action	ACTION_DIALOG_CLOSE;
	
	ImportImageTableModel			imgTabModel;
	private JSGalBuilder			builder;
	private GalleryPanel			galPanel;
	private ImagePreviewPanel		previewPanel;
	private JTable					importTable;
	private JProgressBar			progBar;
	private JComboBox				formatBox;
	private JTextField				qualityField;
	private JCheckBox				progressiveBox;
	private JCheckBox				overwriteBox;
	private JCheckBox				scaleUpBox;
	private JComboBox				useTimestampBox;
	
	private Thread					convThread=null;
	private	boolean					stopThread=false;
	
	protected ImageImportDialog(JSGalBuilder builder, GalleryPanel gp, File imgList[])
	{
		super(builder, I18N.DIALOG_IMGADD_TITLE, true);
		
		setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
		addWindowListener(new WindowAdapter()
		{
			@Override
			public void windowClosing(WindowEvent e)
			{
				closeDialog();
			}		
		});
		this.builder=builder;
		this.galPanel=gp;
		
		initActions();
		
		getContentPane().setLayout(new BorderLayout(4, 4));
	
		// add the preview panel
		previewPanel=new ImagePreviewPanel(new Dimension(200,150));
		getContentPane().add(previewPanel, BorderLayout.EAST);
		
		// create the image list table
		imgTabModel=new ImportImageTableModel(imgList);
		imgTabModel.addTableModelListener(this);
		importTable=new JTable(imgTabModel);
		importTable.getColumnModel().getColumn(1).setCellEditor(new DefaultCellEditor(new JComboBox(I18N.ROTATE_DEGREES)));
		importTable.getColumnModel().getColumn(2).setCellRenderer(new ImportImageTableRenderer());
		
		importTable.getSelectionModel().setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		importTable.getSelectionModel().addListSelectionListener(this);
		importTable.setFillsViewportHeight(true);	// we need this for drag and drop into empty area !!!!
		
		// enable drag n drop
		SwingHelper.enableFileDrop(importTable, new SwingHelper.FileDropListener() 
		{
			@Override
			public boolean canImport(DropLocation location) {	return true;	}	// all locations are valid

			@Override
			public void filesDropped(DropLocation location, File[] files) 
			{
				imgTabModel.add(files);
				
				validateActions();
			}
		});
		
		JPanel	p1=new JPanel(new BorderLayout());
		p1.add(importTable.getTableHeader(), BorderLayout.NORTH);
		p1.add(new JScrollPane(importTable), BorderLayout.CENTER);

		getContentPane().add(p1, BorderLayout.CENTER);
		
		JToolBar	jtb=new JToolBar();
		jtb.add(ACTION_FILE_ADD);
		jtb.add(ACTION_FILE_REMOVE);
		
		getContentPane().add(jtb, BorderLayout.NORTH);
		
		p1=new JPanel(new BorderLayout(4, 4));
		JPanel p2=new JPanel();
		p2.setLayout(new BoxLayout(p2, BoxLayout.Y_AXIS));
		p2.setBorder(BorderFactory.createTitledBorder(I18N.DIALOG_IMGADD_PROGGROUP));
		
		formatBox=SwingHelper.addComboBox(p2, I18N.DIALOG_IMGADD_FORMAT, IMAGE_OUTPUT_FORMATS, BuilderPrefs.GEN_OPTS.format);
		qualityField=SwingHelper.addFloatField(p2, I18N.DIALOG_IMGADD_QUALITY, 4, BuilderPrefs.GEN_OPTS.quality, null);
		progressiveBox=SwingHelper.addCheckBox(p2, I18N.DIALOG_IMGADD_PROGRESSIVE, BuilderPrefs.GEN_OPTS.saveProgressive);
		overwriteBox=SwingHelper.addCheckBox(p2, I18N.DIALOG_IMGADD_OVERWRITE, BuilderPrefs.GEN_OPTS.overwriteExisting);
		scaleUpBox=SwingHelper.addCheckBox(p2, I18N.DIALOG_IMGADD_SCALEUP, BuilderPrefs.GEN_OPTS.scaleIncreaseSize);
		useTimestampBox=SwingHelper.addComboBox(p2, I18N.DIALOG_IMGADD_USETIMESTAMP, I18N.USETIMESTAMP_VALUES, useTS2String(BuilderPrefs.GEN_OPTS.useTimeStampAs));
		
		progBar=new JProgressBar();
		progBar.setStringPainted(true);
		p2.add(Box.createVerticalStrut(8));	
		p2.add(progBar);
		p2.add(Box.createVerticalStrut(4));	
		
		p1.add(new JLabel(I18N.DIALOG_IMGADD_DRAGMSG), BorderLayout.NORTH);
		p1.add(p2, BorderLayout.CENTER);
		
		p2=new JPanel(new FlowLayout(FlowLayout.RIGHT));
		p2.add(new JButton(ACTION_START_CONVERSION));
		p2.add(new JButton(ACTION_STOP_CONVERSION));
		p2.add(Box.createHorizontalStrut(20));
		p2.add(new JButton(ACTION_DIALOG_CLOSE));
		
		p1.add(p2, BorderLayout.SOUTH);
		
		getContentPane().add(p1, BorderLayout.SOUTH);
		
		validateActions();
		pack();
		setLocationRelativeTo(builder);
		setVisible(true);
	}

	private void initActions()
	{
		ACTION_FILE_ADD=new NamedAction(I18N.DIALOG_IMGADD_ADD)
		{ 
			public void actionPerformed(ActionEvent e) { addImages(); } 
		};
		
		ACTION_FILE_REMOVE=new NamedAction(I18N.DIALOG_IMGADD_REMOVE)
		{ 
			public void actionPerformed(ActionEvent e) { removeImages(); } 
		};
		
		ACTION_START_CONVERSION=new NamedAction(I18N.DIALOG_IMGADD_START)
		{ 
			public void actionPerformed(ActionEvent e) { startAddingImages(); } 
		};
		
		ACTION_STOP_CONVERSION=new NamedAction(I18N.DIALOG_IMGADD_STOP)
		{ 
			public void actionPerformed(ActionEvent e) { stopAddingImages(); } 
		};
		
		ACTION_DIALOG_CLOSE=new NamedAction(I18N.DIALOG_IMGADD_CLOSE)
		{ 
			public void actionPerformed(ActionEvent e) { closeDialog(); } 
		};		
	}
	
	void validateActions()
	{
		boolean	running=(convThread!=null);
		
		ACTION_FILE_REMOVE.setEnabled(!running && importTable.getSelectedRow()>=0);		
		ACTION_START_CONVERSION.setEnabled(!running && imgTabModel.size()>0);
		ACTION_STOP_CONVERSION.setEnabled(running);
		ACTION_DIALOG_CLOSE.setEnabled(!running);
	}
	
	void addImages()
	{
		imgTabModel.add(ImageHelper.selectImages(builder, galPanel.getGalleryFile(), I18N.ADD_IMAGES_ADDTITLE, I18N.FILE_CHOOSER_ADD));
		
		validateActions();
	}
	
	void removeImages()
	{
		int	rows[]=importTable.getSelectedRows();
		
		if(rows!=null && rows.length>0)
		{
			for(int i=rows.length-1; i>=0; i--)
			{
				imgTabModel.remove(rows[i]);
			}
			
			imgTabModel.fireTableRowsDeleted(rows[0], rows[rows.length-1]);
		}
		
		validateActions();
	}
	
	void startAddingImages()
	{
		synchronized(this)
		{
			if(convThread==null)
			{
				convThread=new Thread(this);
				convThread.start();
			}
			
			stopThread=false;
		}
		
		validateActions();
	}
	
	void stopAddingImages()
	{
		synchronized(this)
		{
			stopThread=true;
		}
		
		validateActions();
	}
	
	void closeDialog()
	{
		if(convThread==null)
		{
			getUpdatedOpts();
			
			setVisible(false);
			dispose();
		}
	}
	
	private void updateState(int row, ImportImage ii, ImportState newState)
	{
		ii.state=newState;
		imgTabModel.fireTableCellUpdated(row, 2);
	}
	
	private String useTS2String(TimeStampUsage ts)
	{
		switch(ts)
		{
			case NONE:
				return I18N.USETIMESTAMP_NONE;
			case NAME:
				return I18N.USETIMESTAMP_NAME;
			case DESCRIPTION:
				return I18N.USETIMESTAMP_DESCR;
		}
		
		return I18N.USETIMESTAMP_NONE;
	}
	
	private TimeStampUsage string2UseTS(String s)
	{
		if(MiscUtil.equals(s, I18N.USETIMESTAMP_NONE))
			return TimeStampUsage.NONE;
		else if(MiscUtil.equals(s, I18N.USETIMESTAMP_NAME))
			return TimeStampUsage.NAME;
		else if(MiscUtil.equals(s, I18N.USETIMESTAMP_DESCR))
			return TimeStampUsage.DESCRIPTION;
		
		return TimeStampUsage.NONE;
	}
	
	private GenerationOptions getUpdatedOpts()
	{
		GenerationOptions	gopts=BuilderPrefs.GEN_OPTS;
		
		gopts.format=(String)formatBox.getSelectedItem();
		gopts.quality=SwingHelper.getFloatValue(qualityField);
		if(gopts.quality<0.01f)
			gopts.quality=0.01f;
		if(gopts.quality>1f)
			gopts.quality=1f;
		qualityField.setText(Float.toString(gopts.quality));
		gopts.saveProgressive=progressiveBox.isSelected();
		gopts.overwriteExisting=overwriteBox.isSelected();
		gopts.scaleIncreaseSize=scaleUpBox.isSelected();
		gopts.useTimeStampAs=string2UseTS((String)useTimestampBox.getSelectedItem());
		
		return gopts;
	}
	
	public void run()
	{
		boolean				cancelled=false;
		GenerationOptions	gopts=getUpdatedOpts();
		
		progBar.setMinimum(0);

		for(int i=0; i<imgTabModel.size(); i++)
		{
			progBar.setMaximum(imgTabModel.size());		// in case images were added
			
			ImportImage	ii=imgTabModel.get(i);

			progBar.setValue(i);
			String	t=Integer.toString(i+1)+"/"+imgTabModel.size()+" ";	 //$NON-NLS-1$ //$NON-NLS-2$
			progBar.setString(t+ii.imgFile.getName());
			progBar.setToolTipText(t+ii.imgFile.getPath());
			
			if(ii.state.equals(ImportState.INITIAL) || ii.state.equals(ImportState.FAILED))
			{
				try
				{
					updateState(i, ii, ImportState.INPROGRESS);
					
					ImageHelper.addImageToGallery(ii.imgFile, ii.rotate, galPanel, gopts);
					
					updateState(i, ii, ImportState.COMPLETED);
				}
				catch(Exception e)
				{
					ii.exception=e;
					
					updateState(i, ii, ImportState.FAILED);
				}
			}

			synchronized(this)
			{
				if(stopThread)
				{
					// user requested a stop
					convThread=null;
					stopThread=false;

					cancelled=true;
					break;
				}
			}
		}

		if(!cancelled)
		{
			// we're about to exit so clear thread
			synchronized(this)
			{
				convThread=null;
				stopThread=false;
			}
		}
		
		progBar.setValue(imgTabModel.size());
		progBar.setString(null);
		
		validateActions();
	}
	
	@Override
	public void valueChanged(ListSelectionEvent e)
	{
		int	row=importTable.getSelectedRow();
		
		if(row>=0 && row<imgTabModel.size())
		{
			ImportImage	ii=imgTabModel.get(row);
		
			previewPanel.setNewImage(ii.imgFile, null, ii.rotate);
		}
	}

	@Override
	public void tableChanged(TableModelEvent e) 
	{
		valueChanged(null);
	}

	static class ImportImage
	{
		ImportImage(File f)
		{
			imgFile=f;
			state=ImportState.INITIAL;
			rotate=ImageHelper.ROTATE_AUTO;
		}
		
		File			imgFile;
		ImportState		state;
		int				rotate;
		Exception		exception;
	}
	
	static class ImportImageTableRenderer extends DefaultTableCellRenderer
	{
		@Override
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) 
		{
			Object	val=value;
			Icon	icon=null;
			
			if(value instanceof ImportState)
			{
				ImportState	is=(ImportState)value;
				
				val=is.toString();
				icon=is.getIcon();
			}
			
			Component	comp=super.getTableCellRendererComponent(table, val, isSelected, hasFocus, row, column);
			
			setIcon(icon);
			
			return comp;
		}
		
	}
		      	
	static class ImportImageTableModel extends AbstractTableModel
	{
		private Vector<ImportImage>	importImages;
	
		ImportImageTableModel(File images[])
		{
			importImages=new Vector<ImportImage>();
			
			if(images!=null)
			{
				for(File f: images)
					importImages.add(new ImportImage(f));
			}
		}
		
		ImportImage get(int index)
		{
			return importImages.get(index);
		}
		
		int size()
		{
			return importImages.size();
		}
		
		void add(File images[])
		{
			if(images!=null)
			{
				int	inx=importImages.size();
				
				for(File f: images)
					importImages.add(new ImportImage(f));
				
				fireTableRowsInserted(inx, inx+images.length-1);
			}
		}
		
		void add(ImportImage ii)
		{
			int	index=importImages.size();
			
			importImages.add(ii);
			fireTableRowsInserted(index, index);
		}
		
		void remove(int index)
		{
			importImages.remove(index);
			fireTableRowsDeleted(index, index);
		}
		
		@Override
		public Class<?> getColumnClass(int columnIndex) 
		{
			return String.class;
		}

		@Override
		public String getColumnName(int column) 
		{
			return I18N.DIALOG_IMGADD_COL_HEADERS[column];
		}

		@Override
		public int getColumnCount()
		{
			return 3;
		}

		@Override
		public int getRowCount() 
		{
			return importImages.size();
		}

		@Override
		public Object getValueAt(int rowIndex, int columnIndex) 
		{
			ImportImage	ii=importImages.get(rowIndex);
			
			if(ii!=null)
			{
				if(columnIndex==0)
					return ii.imgFile.getName();
				else if(columnIndex==1)
					return I18N.ROTATE_DEGREES[ii.rotate];
				else if(columnIndex==2)
					return ii.state;
			}

			return null;
		}
		
	    @Override
		public boolean isCellEditable(int rowIndex, int columnIndex) 
	    {
	        return (columnIndex==1);
	    }

		@Override
		public void setValueAt(Object value, int rowIndex, int columnIndex)
		{
			if(columnIndex==1)
			{
				ImportImage	ii=importImages.get(rowIndex);
			
				if(ii!=null)
				{
					for(int i=0; i<I18N.ROTATE_DEGREES.length; i++)
					{
						if(I18N.ROTATE_DEGREES[i].equals(value))
						{
							ii.rotate=i;
							fireTableCellUpdated(rowIndex, columnIndex);
							break;
						}
					}
				}
			}
		}
	}
}

enum ImportState 
{
	INITIAL(I18N.STATUS_INITIAL), 
	INPROGRESS(I18N.STATUS_INPROGRESS), 
	COMPLETED(I18N.STATUS_COMPLETED), 
	FAILED(I18N.STATUS_FAILED);
	
	private final String label;
	private ImportState(String label)
	{
		this.label=label;
	}
	
	@Override
	public String toString()
	{
		return label;
	}
	
	public Icon getIcon()
	{
		switch(this)
		{
			case INITIAL:		return I18N.ICON_INITIAL;
			case INPROGRESS:	return I18N.ICON_PROGRESS;
			case COMPLETED:		return I18N.ICON_COMPLETED;
			case FAILED:		return I18N.ICON_FAILED;
		}

		return null;
	}
}